from unittest import TestCase

import pandas as pd

import syndat


class Test(TestCase):

    def test_jsd_zero_int64(self):
        synthetic = pd.DataFrame({
            'feature1': [1, 2, 3, 4, 5],
            'feature2': [10, 11, 12, 13, 14]
        })
        # Create real data
        real = pd.DataFrame({
            'feature1': [6, 7, 8, 9, 10],
            'feature2': [15, 16, 17, 18, 19]
        })
        distribution = syndat.scores.distribution(real, synthetic)
        self.assertEqual(0, distribution)

    def test_jsd_zero_int64_float64(self):
        synthetic = pd.DataFrame({
            'feature1': [1, 2, 3, 4, 5],
            'feature2': [0.1, 0.2, 0.3, 0.4, 0.5]
        })
        # Create real data
        real = pd.DataFrame({
            'feature1': [6, 7, 8, 9, 10],
            'feature2': [0.6, 0.7, 0.8, 0.9, 1.0]
        })
        distribution = syndat.scores.distribution(real, synthetic)
        self.assertEqual(distribution, 0)

    def test_jsd_perfect_int64(self):
        synthetic = pd.DataFrame({
            'feature1': [1, 2, 1, 2, 3],
            'feature2': [11, 12, 13, 14, 15]
        })
        # Create real data
        real = pd.DataFrame({
            'feature1': [1, 2, 1, 2, 3],
            'feature2': [11, 12, 13, 14, 15]
        })
        distribution = syndat.scores.distribution(real, synthetic)
        self.assertEqual(distribution, 100)

    def test_jsd_perfect_int64_and_float64(self):
        synthetic = pd.DataFrame({
            'feature1': [1, 2, 1, 2, 3],
            'feature2': [0.1, 0.2, 0.3, 0.4, 0.5]
        })
        # Create real data
        real = pd.DataFrame({
            'feature1': [1, 2, 1, 2, 3],
            'feature2': [0.1, 0.2, 0.3, 0.4, 0.5]
        })
        distribution = syndat.scores.distribution(real, synthetic)
        self.assertEqual(distribution, 100)

    def test_jsd_different_col_types(self):
        synthetic = pd.DataFrame({
            'feature1': [1, 2, 1, 2, 3],
            'feature2': [1.1, 2.1, 3.1, 4.1, 5.1]
        })
        # Create real data
        real = pd.DataFrame({
            'feature1': [1.2, 2.1, 1.1, 2.1, 3.1],
            'feature2': [1, 2, 3, 4, 5]
        })
        distribution = syndat.scores.distribution(real, synthetic, score=False)

    def test_jsd_negative_int64(self):
        synthetic = pd.DataFrame({
            'feature1': [1, 2, 1, 2, 3],
            'feature2': [1.1, 2.1, 3.1, 4.1, 5.1]
        })
        # Create real data
        real = pd.DataFrame({
            'feature1': [-1, 2, 3, 4, 5],
            'feature2': [1, 2, 3, 4, 5]
        })
        distribution = syndat.scores.distribution(real, synthetic)

    def test_jsd_single_outlier(self):
        synthetic = pd.DataFrame({
            'feature1': [1, 1, 1, 1, 1, 1, 2, 3, 4, 5, 6, 7, 8, 9, 100],
            'feature2': [1, 1, 1, 1, 1, 1,  2, 3, 4, 5, 6, 7, 8, 9, 9],
        })
        # Create real data
        real = pd.DataFrame({
            'feature1': [1, 1, 1, 1, 1, 1, 2, 3, 4, 5, 6, 7, 8, 9, 9],
            'feature2': [1, 1, 1, 1, 1, 1, 2, 3, 4, 5, 6, 7, 8, 9, 100],
        })
        distribution = syndat.scores.distribution(real, synthetic)
        self.assertTrue(distribution < 100)

    def test_jsd_categorical_equal(self):
        synthetic = pd.DataFrame({
            'feature1': ['A', 'B', 'A', 'B', 'C'],
            'feature2': ['X', 'Y', 'Y', 'X', 'Z']
        })
        # Create real data
        real = pd.DataFrame({
            'feature1': ['A', 'B', 'A', 'B', 'C'],
            'feature2': ['X', 'Y', 'Y', 'X', 'Z']
        })
        distribution = syndat.scores.distribution(real, synthetic)
        self.assertEqual(distribution, 100)

    def test_jsd_categorical_different(self):
        synthetic = pd.DataFrame({
            'feature1': ['A', 'B', 'A', 'B', 'C'],
            'feature2': ['X', 'Y', 'Y', 'X', 'Z']
        })
        # Create real data
        real = pd.DataFrame({
            'feature1': ['A', 'B', 'A', 'B', 'D'],
            'feature2': ['X', 'Y', 'Z', 'X', 'W']
        })
        distribution = syndat.scores.distribution(real, synthetic)
        self.assertTrue(distribution < 100)

    def test_jsd_categorical_mixed(self):
        synthetic = pd.DataFrame({
            'feature1': ['A', 'B', 'C', 'D', 'E'],
            'feature2': [1.0, 2.0, 3.0, 4.0, 5.0]
        })
        # Create real data
        real = pd.DataFrame({
            'feature1': ['A', 'B', 'C', 'F', 'G'],
            'feature2': [1.0, 2.0, 3.0, 6.0, 7.0]
        })
        distribution = syndat.scores.distribution(real, synthetic)
        self.assertTrue(distribution < 100)

    def test_jsd_categorical_with_numerical(self):
        synthetic = pd.DataFrame({
            'feature1': ['A', 'B', 'C', 'A', 'B'],
            'feature2': [1.0, 2.0, 3.0, 4.0, 5.0]
        })
        # Create real data
        real = pd.DataFrame({
            'feature1': ['A', 'B', 'C', 'A', 'D'],
            'feature2': [1.0, 2.0, 3.0, 4.0, 6.0]
        })
        distribution = syndat.scores.distribution(real, synthetic)
        self.assertTrue(distribution < 100)

    def test_jsd_categorical_with_nan(self):
        synthetic = pd.DataFrame({
            'feature1': ['A', 'B', 'C', 'D', 'E'],
            'feature2': [1.0, 2.0, 3.0, 4.0, 5.0]
        })
        # Create real data with NaNs
        real = pd.DataFrame({
            'feature1': ['A', 'B', 'C', 'D', None],
            'feature2': [1.0, 2.0, None, 4.0, 5.0]
        })
        distribution = syndat.scores.distribution(real, synthetic)
        self.assertTrue(distribution < 100)

    def test_jsd_categorical_all_nan(self):
        synthetic = pd.DataFrame({
            'feature1': [None, None, None, None, None],
            'feature2': [None, None, None, None, None]
        })
        # Create real data with NaNs
        real = pd.DataFrame({
            'feature1': [None, None, None, None, None],
            'feature2': [None, None, None, None, None]
        })
        distribution = syndat.scores.distribution(real, synthetic)
        self.assertEqual(distribution, 100)



from syndat import domain
from syndat import scores
from syndat import visualization
from syndat import postprocessing

from .visualization import (
    plot_distributions,
    plot_correlations,
    plot_shap_discrimination,
    plot_categorical_feature,
    plot_numerical_feature
)

from .scores import (
    auc,
    jsd,
    discrimination,
    distribution,
    correlation,
)

from .postprocessing import (
    normalize_scale,
    normalize_float_precision,
    assert_minmax,
) 

__all__ = [
    # visualization
    'plot_distributions',
    'plot_correlations',
    'plot_shap_discrimination',
    'plot_categorical_feature',
    'plot_numerical_feature',
    # scores
    'auc',
    'jsd',
    'discrimination',
    'distribution',
    'correlation',
    # postprocessing
    'normalize_scale',
    'normalize_float_precision',
    'assert_minmax',
]



import unittest

import pandas as pd

from syndat.scores import correlation


class TestCorrelation(unittest.TestCase):

    def test_correlation(self):
        real_data = pd.DataFrame({
            'A': [1, 2, 3, 4, 5],
            'B': [5, 4, 3, 2, 1]
        })
        synthetic_data = pd.DataFrame({
            'A': [1, 2, 3, 4, 5],
            'B': [5, 4, 3, 2, 1]
        })
        result = correlation(real_data, synthetic_data)
        self.assertEqual(result, 100, "Correlation score should be 100 for identical datasets")

    def test_correlation_zero(self):
        # positive correlation
        real_data = pd.DataFrame({
            'A': [-1, 1, -2, 2, -3],
            'B': [-2, 2, -4, 4, -6]
        })
        # negative correlation
        synthetic_data = pd.DataFrame({
            'A': [1, -1, 2, -2, 3],
            'B': [-1, 1, -2, 2, -3]
        })
        result = correlation(real_data, synthetic_data)
        self.assertEqual(result, 0)

    def test_correlation_normalized(self):
        real_data = pd.DataFrame({
            'A': [1, 2, 3, 4, 5],
            'B': [5, 4, 3, 2, 1]
        })
        synthetic_data = pd.DataFrame({
            'A': [1, 2, 3, 4, 5],
            'B': [5, 4, 3, 2, 1]
        })
        result = correlation(real_data, synthetic_data, score=False)
        self.assertEqual(0, result, "Normalized correlation score should be 0 for identical datasets")

    def test_correlation_with_categorical(self):
        real_data = pd.DataFrame({
            'risk': ['low', 'low', 'mid', 'high', 'high'],
            'age': [10, 20, 50, 90, 100]
        })
        synthetic_data = pd.DataFrame({
            'risk': ['low', 'low', 'mid', 'high', 'high'],
            'age': [10, 20, 50, 90, 100]
        })
        result = correlation(real_data, synthetic_data)
        self.assertEqual(result, 100, "Correlation score should be 100 for identical datasets with categorical data")

    def test_correlation_with_categorical_more_columns(self):
        real_data = pd.DataFrame({
            'risk': ['low', 'low', 'mid', 'high', 'high'],
            'age': [10, 20, 50, 90, 100],
            'sex': ['male', 'female', 'male', 'female', 'male'],
            'bmi': [21, 22, 23, 24, 25]
        })
        synthetic_data = pd.DataFrame({
            'risk': ['low', 'low', 'mid', 'high', 'high'],
            'age': [10, 20, 50, 90, 100],
            'sex': ['male', 'female', 'male', 'female', 'male'],
            'bmi': [21, 22, 23, 24, 25]
        })
        result = correlation(real_data, synthetic_data)
        self.assertEqual(result, 100, "Correlation score should be 100 for identical datasets with categorical data")

    def test_correlation_with_categorical_extra_category(self):
        real_data = pd.DataFrame({
            'risk': ['low', 'low', 'mid', 'high', 'high', 'very high'],
            'age': [10, 20, 50, 90, 100, 120],
            'sex': ['male', 'female', 'male', 'female', 'male', 'female']
        })
        synthetic_data = pd.DataFrame({
            'risk': ['low', 'low', 'mid', 'high', 'high', 'high'],
            'age': [10, 20, 50, 90, 100, 100],
            'sex': ['male', 'female', 'male', 'female', 'male', 'female']
        })
        result = correlation(real_data, synthetic_data)
        self.assertLessEqual(result, 100, "Correlation score should be 100 for identical "
                                          "datasets with categorical data")

    def test_correlation_with_different_categorical(self):
        real_data = pd.DataFrame({
            # risk of dying of old age
            'risk': ['low', 'low', 'mid', 'high', 'high'],
            'age': [10, 20, 50, 90, 100]
        })
        synthetic_data = pd.DataFrame({
            # risk of having to live for at least another 30 years
            'risk': ['low', 'low', 'mid', 'high', 'high'],
            'age': [100, 90, 50, 20, 10]
        })
        result = correlation(real_data, synthetic_data)
        self.assertLess(result, 100, "Correlation score should be less than 100 for datasets with different "
                                     "categorical data")

    def test_correlation_with_only_categorical(self):
        real_data = pd.DataFrame({
            'sex': ['male', 'female', 'male', 'female', 'male', 'female'],
            'risk': ['high', 'low', 'high', 'low', 'high', 'low'],
        })
        synthetic_data = pd.DataFrame({
            'sex': ['male', 'female', 'male', 'female', 'male', 'female'],
            'risk': ['high', 'high', 'high', 'low', 'high', 'low'],
        })
        result = correlation(real_data, synthetic_data)
        # Adjust the expected result based on encoding and correlation of categorical data
        self.assertLess(result, 100, "Correlation score should be less than 100 for datasets with different "
                                     "categorical data")



import warnings
from typing import Union, List

import logging
import pandas
import pandas as pd
import numpy as np
import scipy.spatial.distance

from sklearn import ensemble
from sklearn.model_selection import cross_val_score, StratifiedKFold
from sklearn.preprocessing import OneHotEncoder

from syndat.domain import AggregationMethod

logger = logging.getLogger(__name__)


def auc(real: pandas.DataFrame, synthetic: pandas.DataFrame, n_folds=5,
        drop_na_threshold=0.9, score: bool = True) -> float:
    """
    Computes the Discrimination Complexity Score / ROC AUC score of a classifier trained to differentiate between real
    and synthetic data.

    :param real: The real data.
    :param synthetic: The synthetic data
    :param n_folds: Number of k folds for cross-validation.
    :param drop_na_threshold: Percentage of non-missing values required of any column
    :param score: Return result in a normalized score in [0,100]. Default is True.
    :return: Differentiation Complexity Score / AUC ROC Score
    """

    warnings.warn(
        "auc is deprecated and will be removed in a future version. Please use discrimination instead.",
        DeprecationWarning,
        stacklevel=2
    )
    return discrimination(real, synthetic, n_folds=n_folds, drop_na_threshold=drop_na_threshold, score=score)


def jsd(real: pd.DataFrame, synthetic: pd.DataFrame, aggregate_results: bool = True,
        aggregation_method: AggregationMethod = AggregationMethod.AVERAGE, score: bool = True,
        n_unique_threshold=10) -> Union[List[float], float]:
    """
    Computes the feature distribution similarity using the Jensen-Shannon distance of real and synthetic data.

    :param real: The real data.
    :param synthetic: The synthetic data.
    :param aggregate_results: Compute a single aggregated score for all features. Default is True.
    :param aggregation_method: How the scores are aggregated. Default is using the median of all feature scores.
    :param score: Return result in a normalized score in [0,100]. Default is True.
    :param n_unique_threshold: Threshold to determine at which number of unique values bins will span over several
    values.
    :return: Distribution Similarity / JSD
    """

    warnings.warn(
        "auc is deprecated and will be removed in a future version. Please use discrimination instead.",
        DeprecationWarning,
        stacklevel=2
    )
    return distribution(real, synthetic, aggregate_results, aggregation_method, score, n_unique_threshold)


def discrimination(real: pandas.DataFrame, synthetic: pandas.DataFrame, n_folds=5,
                   drop_na_threshold=0.9, score: bool = True) -> float:
    """
    Computes the Discrimination Complexity Score / ROC AUC score of a classifier trained to differentiate between real
    and synthetic data.

    :param real: The real data.
    :param synthetic: The synthetic data
    :param n_folds: Number of k folds for cross-validation.
    :param drop_na_threshold: Percentage of non-missing values required of any column
    :param score: Return result in a normalized score in [0,100]. Default is True.
    :return: Differentiation Complexity Score / AUC ROC Score
    """
    # filter out entries that are not inclusive for both datasets
    real_filtered, synthetic_filtered = __filter_rows_with_common_categories(real, synthetic)
    # check for missing values in real data
    real_clean = real_filtered.dropna(thresh=int(drop_na_threshold * len(real_filtered)), axis=1)
    logger.info(f'Dropped {real_clean.shape[1] - real_clean.shape[1]} '
                 f'due to high missingness (threshold is {drop_na_threshold}).')
    real_clean = real_clean.dropna()
    logger.info(f'Removed {len(real) - len(real_clean)} entries due to missing values.')
    # assert that both real and synthetic have same columns
    synthetic_clean = synthetic_filtered[real_clean.columns]
    # one-hot-encode categorical columns
    real_clean_encoded = __encode_categorical(real_clean)
    synthetic_clean_encoded = __encode_categorical(synthetic_clean)
    # train & test the classifier
    x = pd.concat([real_clean_encoded, synthetic_clean_encoded])
    y = np.concatenate((np.zeros(real_clean_encoded.shape[0]), np.ones(synthetic_clean_encoded.shape[0])), axis=None)
    rfc = ensemble.RandomForestClassifier()
    skf = StratifiedKFold(n_splits=n_folds)
    # Calculate cross-validated AUC score
    auc_score = np.mean(cross_val_score(rfc, x, y, cv=skf, scoring='roc_auc'))
    if score:
        return min(0.5, (1 - auc_score)) * 200
    else:
        return auc_score


def distribution(real: pd.DataFrame, synthetic: pd.DataFrame, aggregate_results: bool = True,
                 aggregation_method: AggregationMethod = AggregationMethod.AVERAGE, score: bool = True,
                 n_unique_threshold=10) -> Union[List[float], float]:
    """
    Computes the feature distribution similarity using the Jensen-Shannon distance of real and synthetic data.

    :param real: The real data.
    :param synthetic: The synthetic data.
    :param aggregate_results: Compute a single aggregated score for all features. Default is True.
    :param aggregation_method: How the scores are aggregated. Default is using the median of all feature scores.
    :param score: Return result in a normalized score in [0,100]. Default is True.
    :param n_unique_threshold: Threshold to determine at which number of unique values bins will span over several
    values.
    :return: Distribution Similarity / JSD
    """
    # Initialize a dictionary to store JSD values for each column
    jsd_dict = {}
    # Identify categorical columns
    categorical_columns = real.select_dtypes(include=['object', 'category']).columns
    for col in real:
        # Delete empty cells
        real_wo_missing = real[col].dropna()
        # Check and align data types between real and synthetic data
        col_dtype_real = real[col].dtype
        col_dtype_synthetic = synthetic[col].dtype
        if col_dtype_real != col_dtype_synthetic:
            logger.warning(f'Real data at col {col} is dtype {col_dtype_real} but synthetic is {col_dtype_synthetic}. '
                            f'Evaluation will be done based on the assumed data type of the real data.')
            synthetic[col] = synthetic[col].astype(col_dtype_real)
        # categorical column
        if col in categorical_columns:
            # Handle categorical columns
            real_cat = real[col].astype('category')
            synthetic_cat = synthetic[col].astype('category')
            # Compute frequency distributions
            real_counts = real_cat.value_counts().sort_index()
            synthetic_counts = synthetic_cat.value_counts().sort_index()
            # Fill missing categories with zero
            all_categories = real_counts.index.union(synthetic_counts.index)
            real_binned = real_counts.reindex(all_categories, fill_value=0)
            synthetic_binned = synthetic_counts.reindex(all_categories, fill_value=0)
        # ordinal column
        elif col_dtype_real == "int64" and len(real[col].unique()) <= n_unique_threshold:
            # Handle ordinal columns
            min_value = min(real[col].min(), synthetic[col].min())
            if min_value < 0:
                real[col] += abs(min_value)
                synthetic[col] += abs(min_value)
            real_binned = np.bincount(real[col])
            synthetic_binned = np.bincount(synthetic[col])
        # numerical column
        else:
            # Handle numerical columns
            n_bins = np.histogram_bin_edges(real_wo_missing, bins='auto')
            real_binned = np.bincount(np.digitize(real_wo_missing, n_bins))
            synthetic_binned = np.bincount(np.digitize(synthetic[col], n_bins))
        # One array might be shorter than the other; fill missing bins with zero
        if len(real_binned) != len(synthetic_binned):
            padding_size = abs(len(real_binned) - len(synthetic_binned))
            if len(real_binned) > len(synthetic_binned):
                synthetic_binned = np.pad(synthetic_binned, (0, padding_size))
            else:
                real_binned = np.pad(real_binned, (0, padding_size))
        # Compute JSD
        jsd_value = scipy.spatial.distance.jensenshannon(real_binned, synthetic_binned, 2)
        jsd_dict[col] = jsd_value
    if not aggregate_results:
        return jsd_dict
    if aggregation_method == AggregationMethod.AVERAGE:
        jsd_aggregated = np.mean(list(jsd_dict.values()))
    elif aggregation_method == AggregationMethod.MEDIAN:
        jsd_aggregated = np.median(list(jsd_dict.values()))
    if not score:
        return jsd_aggregated
    else:
        return int((1 - jsd_aggregated) * 100)


def correlation(real: pd.DataFrame, synthetic: pd.DataFrame, score=True) -> float:
    """
    Computes the correlation similarity of real and synthetic data.

    Filters out rows with categories not present in both datasets, encodes categorical columns,
    and computes the correlation matrix.

    The goal is to ensure that only comparable data is used in the correlation calculation.

    :param real: The real data.
    :param synthetic: The synthetic data.
    :param score: Return result in a normalized score in [0,100]. Default is True.
    :return: Correlation score / Norm Quotient
    """
    # why do we do this? while it is expected that categorical columns of both data sets contain the same categories,
    # it is entirely possible that this is not the case. If this happens, we omit this extra category from computation
    # to avoid possible crashes in correlation computation.
    real_filtered, synthetic_filtered = __filter_rows_with_common_categories(real, synthetic)
    # Encode categorical columns
    real_encoded = __encode_categorical(real_filtered)
    synthetic_encoded = __encode_categorical(synthetic_filtered)
    # Compute numerical correlation only
    real_numerical = real_encoded.select_dtypes(include=[np.number])
    synthetic_numerical = synthetic_encoded.select_dtypes(include=[np.number])
    # Compute correlation matrices
    corr_real = real_numerical.corr()
    corr_synthetic = synthetic_numerical.corr()
    # Remove one-hot-encoded categories from one dimension - otherwise we compute correlations within the same column
    # which would distort results
    one_hot_encoded_columns = list(set(real_encoded.columns) - set(real.columns))
    # beware of special case: only categorical values. In this case we cant drop all columns because the dataframe
    # would be empty afterwards TODO: handle this more gracefully > differentiate between encodings of different cats
    if not corr_real.drop(columns=one_hot_encoded_columns).empty:
        corr_real = corr_real.drop(columns=one_hot_encoded_columns)
        corr_synthetic = corr_synthetic.drop(columns=one_hot_encoded_columns)
    # now compute correlation matrices
    norm_diff = np.linalg.norm(corr_real - corr_synthetic)
    norm_real = np.linalg.norm(corr_real)
    norm_quotient = norm_diff / norm_real
    if score:
        # Normalize to score in [0, 100]
        if norm_quotient > 1:
            return 0
        return (1 - max(norm_quotient, 0)) * 100
    else:
        return norm_quotient


def __filter_rows_with_common_categories(real: pd.DataFrame, synthetic: pd.DataFrame) -> (pd.DataFrame, pd.DataFrame):
    """
    Filters out rows with categories that are not present in both datasets.

    Only keeps rows with categories that are present in both datasets.
    This ensures that both datasets are comparable in terms of categorical data.

    :param real: The real data.
    :param synthetic: The synthetic data.
    :return: Filtered real and synthetic datasets with only common categories in rows.
    """
    # Identify categorical columns
    real_categorical_cols = real.select_dtypes(include=['object', 'category']).columns
    synthetic_categorical_cols = synthetic.select_dtypes(include=['object', 'category']).columns
    # Identify common categorical columns
    common_categorical_cols = set(real_categorical_cols) & set(synthetic_categorical_cols)
    # Filter rows with common categories in each column
    for col in common_categorical_cols:
        real_categories = set(real[col].unique())
        synthetic_categories = set(synthetic[col].unique())
        common_categories = real_categories & synthetic_categories
        if len(real_categories - common_categories) > 0:
            logger.warning(
                f"Categories {real_categories - common_categories} in column '{col}' are in real data but not in "
                f"synthetic data. They will not be considered in the score computation.")
        # Filter rows to keep only common categories
        real = real[real[col].isin(common_categories)]
        synthetic = synthetic[synthetic[col].isin(common_categories)]
    return real, synthetic


def __encode_categorical(data: pd.DataFrame) -> pd.DataFrame:
    """
    Encodes categorical columns in the dataframe using one-hot encoding.

    Converts categorical variables into a format suitable for correlation computation.

    :param data: The dataframe with potential categorical columns.
    :return: DataFrame with categorical columns one-hot encoded.
    """
    categorical_cols = data.select_dtypes(include=['object', 'category']).columns
    if len(categorical_cols) == 0:
        return data
    # Initialize encoder without dropping any category
    encoder = OneHotEncoder(sparse_output=False)
    encoded_data = data.copy()
    for col in categorical_cols:
        encoded_col = encoder.fit_transform(data[[col]])
        encoded_df = pd.DataFrame(encoded_col, columns=encoder.get_feature_names_out([col]), index=data.index)
        encoded_data = encoded_data.drop(col, axis=1)
        encoded_data = pd.concat([encoded_data, encoded_df], axis=1)
    return encoded_data



import pandas as pd
import numpy as np
import syndat

import unittest


class Test(unittest.TestCase):

    def setUp(self):
        # Create synthetic data with numerical features
        self.synthetic_data = pd.DataFrame({
            'feature1': np.random.rand(100),
            'feature2': np.random.rand(100)
        })

        # Create real data with numerical features
        self.real_data = pd.DataFrame({
            'feature1': np.random.rand(100),
            'feature2': np.random.rand(100)
        })

        # Create real and synthetic data with categorical features
        self.real_data_cat = pd.DataFrame({
            'feature1': np.random.choice(['A', 'B', 'C'], size=100),
            'feature2': np.random.choice(['X', 'Y'], size=100)
        })

        self.synthetic_data_cat = pd.DataFrame({
            'feature1': np.random.choice(['A', 'B', 'C'], size=100),
            'feature2': np.random.choice(['X', 'Y'], size=100)
        })

    def preprocess_categorical_data(self, real_data, synthetic_data):
        # Convert categorical features to one-hot encoding
        real_data_encoded = pd.get_dummies(real_data)
        synthetic_data_encoded = pd.get_dummies(synthetic_data)

        # Align the columns
        real_data_encoded, synthetic_data_encoded = real_data_encoded.align(synthetic_data_encoded, join='left', axis=1,
                                                                            fill_value=0)

        return real_data_encoded, synthetic_data_encoded

    def test_auc_score(self):
        auc_score = syndat.scores.discrimination(self.real_data, self.synthetic_data)
        self.assertTrue(isinstance(auc_score, float))
        self.assertGreaterEqual(auc_score, 0.0)
        self.assertLessEqual(auc_score, 100.0)

    def test_auc_score_normalized(self):
        auc_score = syndat.scores.discrimination(self.real_data, self.synthetic_data)
        self.assertTrue(isinstance(auc_score, float))
        self.assertGreaterEqual(auc_score, 0.0)
        self.assertLessEqual(auc_score, 100.0)

    def test_auc_score_with_missing_values(self):
        # Introduce missing values in real data
        self.real_data.iloc[::10, 0] = np.nan  # 10% missing data
        auc_score = syndat.scores.discrimination(self.real_data, self.synthetic_data)
        self.assertTrue(isinstance(auc_score, float))
        self.assertGreaterEqual(auc_score, 0.0)
        self.assertLessEqual(auc_score, 100.0)

    def test_auc_score_with_missing_values_drop_col(self):
        # Introduce missing values in real data
        self.real_data.iloc[::2, 0] = np.nan  # 50% missing data -> col drop
        auc_score = syndat.scores.discrimination(self.real_data, self.synthetic_data)
        self.assertTrue(isinstance(auc_score, float))
        self.assertGreaterEqual(auc_score, 0.0)
        self.assertLessEqual(auc_score, 100.0)

    def test_auc_score_with_custom_folds(self):
        auc_score = syndat.scores.discrimination(self.real_data, self.synthetic_data, n_folds=5)
        self.assertTrue(isinstance(auc_score, float))
        self.assertGreaterEqual(auc_score, 0.0)
        self.assertLessEqual(auc_score, 100.0)

    def test_auc_score_with_categorical_data(self):
        auc_score = syndat.scores.discrimination(self.real_data_cat, self.synthetic_data_cat)
        self.assertTrue(isinstance(auc_score, float))
        self.assertGreaterEqual(auc_score, 0.0)
        self.assertLessEqual(auc_score, 100.0)



import pandas as pd
import numpy as np
from sklearn.ensemble import RandomForestClassifier
from sklearn.model_selection import train_test_split
from sklearn.metrics import roc_auc_score
import matplotlib.pyplot as plt
import seaborn as sns

def normalize_scale(real_df: pd.DataFrame, synthetic_df: pd.DataFrame) -> pd.DataFrame:
    """
    Scales the columns in the synthetic DataFrame to match the scale (min and max values) of the corresponding columns in the real DataFrame.

    Parameters:
    real_df (pd.DataFrame): The real dataset used as the scaling reference.
    synthetic_df (pd.DataFrame): The synthetic dataset to be scaled.

    Returns:
    pd.DataFrame: The scaled synthetic dataset with columns adjusted to the real dataset's scale.
    """
    # Create a copy of the synthetic dataframe to avoid modifying the original one
    scaled_synthetic_df = synthetic_df.copy()

    # Iterate over each column in the synthetic dataframe
    for column in synthetic_df.columns:
        # Check if the column is of floating-point type in both real and synthetic data
        if np.issubdtype(synthetic_df[column].dtype, np.floating):
            # Find the min and max values in the real data for this column
            real_min = real_df[column].min()
            real_max = real_df[column].max()
            
            # Find the min and max values in the synthetic data for this column
            synthetic_min = synthetic_df[column].min()
            synthetic_max = synthetic_df[column].max()
            
            # Scale the synthetic data to match the min/max of the real data
            scaled_synthetic_df[column] = ((synthetic_df[column] - synthetic_min) / (synthetic_max - synthetic_min)) * (real_max - real_min) + real_min

    return scaled_synthetic_df

def assert_minmax(real: pd.DataFrame, synthetic: pd.DataFrame, method: str = 'clip') -> pd.DataFrame:
    """
    Postprocess the synthetic data by either deleting records that fall outside the min-max range of the real data,
    or adjusting them to fit within the range. Also normalizes -0.0 to 0.0 to avoid plotting issues.

    Parameters:
    real (pd.DataFrame): The real dataset.
    synthetic (pd.DataFrame): The synthetic dataset.
    method (str): The method to apply. 'delete' to remove records, 'clip' to adjust them.

    Returns:
    pd.DataFrame: The postprocessed synthetic dataset.
    """
    # Normalize -0.0 to 0.0 in synthetic data
    synthetic = synthetic.apply(lambda col: col.map(lambda x: 0.0 if x == -0.0 else x))
    
    # Iterate over each column in the synthetic DataFrame
    for column in synthetic.columns:
        if column in real.columns:
            # Get the min and max of the column in the real data
            min_val = real[column].min()
            max_val = real[column].max()
            
            if method == 'delete':
                # Filter the synthetic DataFrame to keep only rows within the min-max range
                synthetic = synthetic[(synthetic[column] >= min_val) & (synthetic[column] <= max_val)]
            elif method == 'clip':
                # Clip the values to be within the min-max range
                synthetic[column] = synthetic[column].clip(lower=min_val, upper=max_val)

    return synthetic

def normalize_float_precision(real: pd.DataFrame, synthetic: pd.DataFrame) -> pd.DataFrame:
    """
    Postprocess the synthetic data to match the precision or step size found in the real data for float columns.

    This function identifies columns in the real dataset that have float data types and determines the precision
    or step size (e.g., 1.0, 0.5, 0.1) used in those columns. It then rounds the corresponding columns in the 
    synthetic dataset to match this detected precision or step size.

    Parameters:
    real (pd.DataFrame): The real dataset containing float columns.
    synthetic (pd.DataFrame): The synthetic dataset that needs to be adjusted to match the precision of the real data.

    Returns:
    pd.DataFrame: The synthetic dataset with float columns rounded to match the precision or step size of the real data.
    """
    # Select float columns from the real dataset
    float_columns = real.select_dtypes(include='float').columns

    for col in float_columns:
        if col in synthetic.columns:
            # Get the unique values from the real data column, excluding NaN
            unique_values = real[col].dropna().unique()
            
            # Calculate the differences between the unique sorted values
            unique_diffs = np.diff(np.sort(unique_values))
            
            # If the unique values are all the same, continue to the next column
            if len(unique_diffs) == 0:
                continue
            
            # Find the smallest non-zero difference (the step size)
            step_size = np.min(unique_diffs[unique_diffs > 0])
            
            # Round the synthetic column to the nearest multiple of the step size
            synthetic[col] = np.round(synthetic[col] / step_size) * step_size

    return synthetic


import io
import os

from setuptools import setup, find_packages

DESCRIPTION = 'A library for evaluation & visualization of synthetic data.'

here = os.path.abspath(os.path.dirname(__file__))

try:
    with io.open(os.path.join(here, 'README.md'), encoding='utf-8') as f:
        long_description = '\n' + f.read()
except FileNotFoundError:
    long_description = DESCRIPTION

setup(
    name='syndat',
    version='v0.10.0',
    packages=find_packages(exclude=["tests", "*.tests", "*.tests.*", "tests.*"]),
    url='https://github.com/SCAI-BIO/syndat',
    license='MIT',
    author='Tim Adams',
    author_email='tim.adams@scai.fraunhofer.de',
    description=DESCRIPTION,
    long_description=long_description,
    long_description_content_type='text/markdown',
    install_requires=[
        'pandas~=2.1',
        'numpy~=1.26',
        'scipy~=1.11',
        'scikit-learn~=1.5',
        'matplotlib~=3.8',
        'seaborn~=0.13',
        'shap~=0.42.0',
        'setuptools>=70.0.0'
    ],
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3.11',
    ],
    include_package_data=True,  
    python_requires='>=3.9',  
    keywords='synthetic-data, data-quality, data-visualization',
    project_urls={
        'Documentation': 'https://github.com/SCAI-BIO/syndat#readme',
        'Source': 'https://github.com/SCAI-BIO/syndat',
        'Tracker': 'https://github.com/SCAI-BIO/syndat/issues',
    },
)



import pandas
import matplotlib
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import seaborn as sns
from pandas.plotting import table
from sklearn.model_selection import train_test_split  
from sklearn.ensemble import RandomForestClassifier  
from sklearn.metrics import roc_auc_score  
import shap  # Added


def plot_distributions(real: pandas.DataFrame, synthetic: pandas.DataFrame, store_destination: str) -> None:
    """
    Plots violin plots (numeric features) or bar charts (categorical features) together with their summary statistics.

    :param real: The real data
    :param synthetic: The synthetic data
    :param store_destination: Path to the folder where the results should be stored.
    """
    for column_name in real.columns:
        matplotlib.use('Agg')
        real_col = real[column_name].to_numpy()
        synthetic_col = synthetic[column_name].to_numpy()
        plt.figure()
        plt.title(column_name)
        patient_types = np.concatenate([np.zeros(real_col.size), np.ones(synthetic_col.size)])
        df = pd.DataFrame(data={"type": np.where(patient_types == 0, "real", "synthetic"),
                                "value": np.concatenate([real_col, synthetic_col])})
        if real_col.dtype == str or real_col.dtype == object:
            ax = sns.countplot(data=df, x="value", hue="type", order=df['value'].value_counts().index)
        elif np.sum(real_col) % 1 == 0 and np.max(real_col) < 10:
            ax = sns.countplot(data=df, x="value", hue="type")
        else:
            if len(real_col) != len(synthetic_col):
                # Determine the length of the longest column
                max_length = max(len(real_col), len(synthetic_col))
                # Fill the shorter column with NaN values to match the length of the longest column
                real_col = np.pad(real_col, pad_width=(0, max_length - len(real_col)), mode='constant',
                                  constant_values=np.nan)
                synthetic_col = np.pad(synthetic_col, pad_width=(0, max_length - len(synthetic_col)), mode='constant',
                                       constant_values=np.nan)
            # create df with now equal values
            df = pd.DataFrame(data={"real": real_col, "synthetic": synthetic_col})
            ax = sns.violinplot(data=df)
            # remove y-labels as they are redundant with the table headers
            ax.set_xticks([])
            table(ax, df.describe().round(2), loc='bottom', colLoc='center', bbox=[0, -0.55, 1, 0.5],
                  colWidths=[.5, .5])
        fig = ax.get_figure()
        matplotlib.pyplot.close()
        fig.savefig(store_destination + "/" + column_name + '.png', bbox_inches="tight")


def plot_correlations(real: pandas.DataFrame, synthetic: pandas.DataFrame, store_destination: str) -> None:
    """
    Plots correlation matrices for real and synthetic features in form of heatmaps.

    :param real: The real data
    :param synthetic: The synthetic data
    :param store_destination: Path to the folder where the results should be stored.
    """
    names = ["real_corr", "syntehtic_corr"]
    for idx, patient_type in enumerate([real, synthetic]):
        plt.figure()
        plt.title("Correlation")
        ax = sns.heatmap(patient_type.corr(), vmin=-1, vmax=1)
        fig = ax.get_figure()
        fig.savefig(store_destination + "/" + names[idx] + '.png', bbox_inches="tight")

def plot_shap_discrimination(real: pandas.DataFrame, synthetic: pandas.DataFrame) -> None:
    """
    Trains a Random Forest Classifier to discriminate between real and synthetic data and plots SHAP summary values.

    :param real: The real data
    :param synthetic: The synthetic data
    """
    # Assuming 'real' and 'synthetic' are your datasets and are pandas DataFrames
    # Add a label column to each dataset
    real['label'] = 1
    synthetic['label'] = 0

    # Combine datasets
    combined_data = pd.concat([real, synthetic])

    # Separate features and labels
    X = combined_data.drop('label', axis=1)
    y = combined_data['label']

    # Split data into training and test sets
    X_train, X_test, y_train, y_test = train_test_split(X, y, test_size=0.2, random_state=42, stratify=y)

    # Initialize and train the Random Forest Classifier
    rfc = RandomForestClassifier(n_estimators=100, random_state=42)
    rfc.fit(X_train, y_train)

    # Predict probabilities and calculate AUC score
    y_pred_proba = rfc.predict_proba(X_test)[:, 1]
    auc_score = roc_auc_score(y_test, y_pred_proba)
    print(f'AUC Score: {auc_score}')

    # Compute SHAP values
    explainer = shap.TreeExplainer(rfc)
    shap_values = explainer.shap_values(X_test)

    # Plot SHAP summary
    shap.summary_plot(shap_values[:, :, 1], X_test)


def plot_categorical_feature(feature: str, real_data: pandas.DataFrame, synthetic_data: pandas.DataFrame) -> None:
    """
    Plots count plots for a categorical feature from both real and synthetic datasets.

    :param feature: The feature to be plotted
    :param real_data: The real data
    :param synthetic_data: The synthetic data
    """
    plt.figure(figsize=(14, 6))

    # Plot for the real dataset
    plt.subplot(1, 2, 1)
    sns.countplot(x=feature, data=real_data, color='blue')
    plt.xlabel(feature)
    plt.ylabel('Frequency')
    plt.title(f'Real Data - {feature}')
    plt.xticks(rotation=90)

    # Plot for the synthetic dataset
    plt.subplot(1, 2, 2)
    sns.countplot(x=feature, data=synthetic_data, color='orange')
    plt.xlabel(feature)
    plt.ylabel('Frequency')
    plt.title(f'Synthetic Data - {feature}')
    plt.xticks(rotation=90)

    plt.tight_layout()
    plt.show()


def plot_numerical_feature(feature: str, real_data: pandas.DataFrame, synthetic_data: pandas.DataFrame) -> None:
    """
    Plots violin plots for a numerical feature from both real and synthetic datasets and displays their summary statistics.

    :param feature: The feature to be plotted
    :param real_data: The real data
    :param synthetic_data: The synthetic data
    """
    # Calculate summary statistics
    def get_summary_stats(data, feature):
        return {
            'Mean': data[feature].mean(),
            'Median': data[feature].median(),
            'Std Dev': data[feature].std(),
            'Min': data[feature].min(),
            'Max': data[feature].max()
        }
    
    real_stats = get_summary_stats(real_data, feature)
    synthetic_stats = get_summary_stats(synthetic_data, feature)
    
    # Create summary statistics DataFrame
    stats_df = pd.DataFrame({
        'Statistic': ['Mean', 'Median', 'Std Dev', 'Min', 'Max'],
        'Real Data': [real_stats['Mean'], real_stats['Median'], real_stats['Std Dev'], real_stats['Min'], real_stats['Max']],
        'Synthetic Data': [synthetic_stats['Mean'], synthetic_stats['Median'], synthetic_stats['Std Dev'], synthetic_stats['Min'], synthetic_stats['Max']]
    })
    
    plt.figure(figsize=(14, 8))

    # Compute the combined range for x-axis limits
    min_value = min(real_data[feature].min(), synthetic_data[feature].min())
    max_value = max(real_data[feature].max(), synthetic_data[feature].max())

    # Plot for the real dataset
    plt.subplot(2, 2, 1)
    sns.violinplot(x=real_data[feature], color='blue')
    plt.xlabel(feature)
    plt.ylabel('Density')
    plt.title(f'Real Data - {feature}')
    plt.xlim(min_value, max_value)

    # Plot for the synthetic dataset
    plt.subplot(2, 2, 2)
    sns.violinplot(x=synthetic_data[feature], color='orange')
    plt.xlabel(feature)
    plt.ylabel('Density')
    plt.title(f'Synthetic Data - {feature}')
    plt.xlim(min_value, max_value)

    # Display summary statistics table
    plt.subplot(2, 1, 2)
    plt.axis('off')
    table = plt.table(cellText=stats_df.values,
                      colLabels=stats_df.columns,
                      rowLabels=stats_df['Statistic'],
                      cellLoc='center',
                      loc='center',
                      bbox=[0.0, -0.5, 1.0, 0.4])
    table.auto_set_font_size(False)
    table.set_fontsize(10)
    table.scale(1.2, 1.2)  # Adjust the size of the table

    plt.tight_layout()
    plt.show()




